#include "Main.h"
#include "Menu.h"

#include <chrono>
#include <iostream>
#include <SDL.h>
#include <glm/vec3.hpp>
#include <Windows.h>

void Input(SDL_Keycode code, Sandbox& sandbox)
{
	switch (code)
	{
		// Menu
	case SDLK_m: sandbox.MenuOpen = !sandbox.MenuOpen; break;

		// Drawsize
	case SDLK_KP_PLUS:  if (sandbox.drawsize < MAXDRAWSIZE) sandbox.drawsize += 2; break;
	case SDLK_KP_MINUS: if (sandbox.drawsize > 1)			sandbox.drawsize -= 2; break;

		// Speed
	case SDLK_1: sandbox.Speed = 1 << 0; break;
	case SDLK_2: sandbox.Speed = 1 << 1; break;
	case SDLK_3: sandbox.Speed = 1 << 2; break;
	case SDLK_4: sandbox.Speed = 1 << 3; break;
	case SDLK_5: sandbox.Speed = 1 << 4; break;
	case SDLK_6: sandbox.Speed = 1 << 5; break;
	case SDLK_7: sandbox.Speed = 1 << 6; break;
	case SDLK_8: sandbox.Speed = 1 << 7; break;
	case SDLK_9: sandbox.Speed = 1 << 8; break;
	case SDLK_0: sandbox.Speed = 0; break;
	default: break;
	}
}

void PrintDebug(HANDLE console, Sandbox sandbox, float dt)
{	
	SetConsoleCursorPosition(console, { 0, 0 });

	printf("FPS: %i          \n", (int)(1.f / dt));
	printf("Pixels: %i       \n", sandbox.cells.size());
	printf("Speed: %i        \n\n", sandbox.Speed);

	std::string mat;
	switch (sandbox.draw)
	{
	case Types::Air:	mat = "Delete"; break;
	case Types::Sand:	mat = "Sand  "; break;
	case Types::Water:	mat = "Water "; break;
	case Types::Lava:	mat = "Lava  "; break;
	case Types::Stone:	mat = "Stone "; break;
	case Types::Clone:	mat = "Clone "; break;
	case Types::Block:	mat = "Block "; break;
	case Types::Ice:	mat = "Ice   "; break;
	case Types::Cloud:	mat = "Cloud "; break;
	case Types::Fire:	mat = "Fire  "; break;
	case Types::Oil:	mat = "Oil   "; break;
	case Types::Powder:	mat = "Powder"; break;
	case Types::Acid:	mat = "Acid  "; break;
	}
	printf("Draw size: %i   \n", sandbox.drawsize);
	printf("Currently material: %s\n\n\n", mat.c_str());


	printf(
		"Keybinds:\n"
		"M = Open/Close menu\n\n"

		"Keypad Plus  = Increase drawsize\n"
		"Keypad Minus = Decrease drawsize\n\n"

		"Numbers 1-9 = Change Speed\n"
		"Numbers 0   = Pause Game\n");
}

int main(int argc, char** argv)
{
	Sandbox sandbox;
	for (int i = 0; i < SCRHEIGHT * SCRWIDTH; i++)
	{
		sandbox.grid[i] = nullptr;
	}

	// init border
	for (int y = 0; y < SCRHEIGHT; y++)
	{
		for (int x = 0; x < SCRWIDTH; x++)
		{
			if(x < BORDERSIZE || x > SCRWIDTH - BORDERSIZE || y < BORDERSIZE || y > SCRHEIGHT - BORDERSIZE)
				sandbox.cells.push_back(new Update(sandbox, Types::Block, glm::ivec2(x, y), true));
		}
	}

	Menu menu;
	menu.InitMenu(sandbox);

	bool ExitApp = false;
	SDL_Event Event;
	int FrameCount = 0;
	float delta_time = 0.0001f;

	SDL_Init(SDL_INIT_VIDEO);

	SDL_Window* screen = SDL_CreateWindow("Sandbox", SDL_WINDOWPOS_CENTERED_DISPLAY(DISPLAY), SDL_WINDOWPOS_CENTERED_DISPLAY(DISPLAY), SCRWIDTH, SCRHEIGHT, SDL_WINDOW_BORDERLESS);
	SDL_Renderer* renderer = SDL_CreateRenderer(screen, -1, 0);
	
	HANDLE console = GetStdHandle(STD_OUTPUT_HANDLE);

	while (!ExitApp)
	{
		auto start = std::chrono::steady_clock::now();

		// if we've paused we don't want to teleport everything ahead by a lot
		if (delta_time > 0.1f)
			delta_time = 0.1f;

		SDL_PollEvent(&Event);				

		// mouse input
		{
			int MouseX;
			int MouseY;
			Uint32 state = SDL_GetMouseState(&MouseX, &MouseY);

			if (!sandbox.MenuOpen || MouseX < SCRWIDTH - MENUSIZE)
			{
				if ((state == SDL_BUTTON_LEFT && sandbox.draw == Types::Air))
				{
					int offset = (int)floor((float)sandbox.drawsize / 2.f);
					int MenuMargin = sandbox.MenuOpen ? SCRWIDTH - MENUSIZE : SCRWIDTH;

					for (int y = MouseY - offset; y <= MouseY + offset; y++)
					{
						for (int x = MouseX - offset; x <= MouseX + offset; x++)
						{
							if (x >= 0 && x < MenuMargin && y >= 0 && y < SCRHEIGHT)
							{
								// Delete pixel
								Update* DeletedCell = sandbox.grid[y * SCRWIDTH + x];
								if (DeletedCell != nullptr)
								{
									sandbox.cells.erase(std::find(sandbox.cells.begin(), sandbox.cells.end(), DeletedCell));
									sandbox.grid[DeletedCell->pos.y * SCRWIDTH + DeletedCell->pos.x] = nullptr;
									delete DeletedCell;
								}
							}
						}
					}
				}
				else if (state == SDL_BUTTON_LEFT)
				{
					int offset = (int)floor((float)sandbox.drawsize / 2.f);

					for (int y = MouseY - offset; y <= MouseY + offset; y++)
					{
						for (int x = MouseX - offset; x <= MouseX + offset; x++)
						{
							if (x >= 0 && x < SCRWIDTH && y >= 0 && y < SCRHEIGHT)
							{
								if (sandbox.grid[y * SCRWIDTH + x] == nullptr)
								{
									sandbox.cells.push_back(new Update(sandbox, sandbox.draw, glm::ivec2(x, y), true));
								}
							}
						}
					}
				}
			}
			else if (state == SDL_BUTTON_LEFT) {
				ExitApp = menu.CheckMouse(sandbox, glm::ivec2(MouseX, MouseY));
			}
		}

		// Update cells
		int size = sandbox.cells.size();
		if (sandbox.Speed != 0)
		{
			for (int i = 0; i < size; i++)
			{
				sandbox.cells[i]->UpdateCell(sandbox);
			}
		}

		// Delete cells if needed
		size = sandbox.cells.size();
		for (int i = size - 1; i >= 0; i--)
		{
			if (sandbox.cells[i]->QueueDelete)
			{
				Update* cell = sandbox.cells[i];

				sandbox.cells.erase(std::find(sandbox.cells.begin(), sandbox.cells.end(), cell));
				sandbox.grid[cell->pos.y * SCRWIDTH + cell->pos.x] = nullptr;
				delete cell;
			}
		}


		// Render cells
		if (sandbox.Speed == 0 || FrameCount % sandbox.Speed == 0)
		{
			int margin = SCRWIDTH - MENUSIZE;

			for (auto cell : sandbox.cells)
			{
				if (!sandbox.MenuOpen || cell->pos.x < margin)
				{
					glm::ivec3 color = cell->RenderCell();

					//draw the pixel on screen
					SDL_SetRenderDrawColor(renderer, (int)color.x, (int)color.y, (int)color.z, 0);
					SDL_RenderDrawPoint(renderer, cell->pos.x, cell->pos.y);
				}
			}

			if (sandbox.MenuOpen)
			{
				for (auto cell : sandbox.Menu)
				{
					glm::ivec3 color = cell->RenderCell();

					//draw the pixel on screen
					SDL_SetRenderDrawColor(renderer, (int)color.x, (int)color.y, (int)color.z, 0);
					SDL_RenderDrawPoint(renderer, cell->pos.x, cell->pos.y);
				}
			}


			//present the renderer to the window
			SDL_RenderPresent(renderer);
			SDL_SetRenderDrawColor(renderer, 0, 0, 0, 0);
			SDL_RenderClear(renderer);
		}

		switch (Event.type)
		{
		case SDL_QUIT:
			ExitApp = true;
			break;
		case SDL_KEYDOWN:
			if (Event.key.keysym.sym == SDLK_ESCAPE) {
				ExitApp = true;
			}
			else Input(Event.key.keysym.sym, sandbox);
		default:
			break;
		}
		FrameCount++;

		if (FPSCAP != -1)
		{
			auto end = std::chrono::steady_clock::now();
			std::chrono::duration<float> elapsed_chrono_time = end - start;
			delta_time = elapsed_chrono_time.count();

			float time = 1.f / (float)(FPSCAP * sandbox.Speed);
			if (delta_time < time)
			{
				Sleep(DWORD((time - delta_time) * 1000.f));
				delta_time = time;
			}
		}
		else
		{
			auto end = std::chrono::steady_clock::now();
			std::chrono::duration<float> elapsed_chrono_time = end - start;
			delta_time = elapsed_chrono_time.count();
		}
		
		if (SHOWDEBUG)
			PrintDebug(console, sandbox, delta_time);
	}

	// Quit game
	SDL_Quit();
	for (auto cell : sandbox.cells)
	{
		delete cell;
	}
	for (auto cell : sandbox.Menu)
	{
		delete cell;
	}

	return 0;
}